import abc
import base64
from typing import Dict
from dataclasses import dataclass


class Auth(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_headers(self) -> Dict[str, str]:
        raise NotImplementedError()


@dataclass
class BasicAuth(Auth):
    username: str
    password: str

    def get_headers(self):
        raw_token = f"{self.username}:{self.password}"
        token = base64.b64encode(raw_token.encode("ascii")).decode("ascii")
        return {"Authorization": f"Basic {token}"}



import asyncio
import base64
import json
from functools import wraps
from io import BytesIO
from json import JSONDecodeError
import logging
from pathlib import Path
from typing import Any, Tuple
from urllib.parse import urlparse, urlunparse
from zipfile import ZIP_DEFLATED, ZipFile

import httpx

from wapiti_arsenic import errors, constants

log = logging.getLogger(__name__)


def wrap_screen(data):
    """
    Data returned from a webdriver may contain a screen, which is a base64
    encoded PNG of the browser screen. This is a massive string and will make
    logging useless, so we wrap it in a BytesIO.
    """
    if (
        isinstance(data, dict)
        and "value" in data
        and isinstance(data["value"], dict)
        and "screen" in data["value"]
        and data["value"]["screen"]
    ):
        data["value"]["screen"] = BytesIO(base64.b64decode(data["value"]["screen"]))


def unwrap(value):
    if isinstance(value, dict) and (
        "ELEMENT" in value or constants.WEB_ELEMENT in value
    ):
        wrapped_id = value.get("ELEMENT", None)
        if wrapped_id:
            return value["ELEMENT"]
        else:
            return value[constants.WEB_ELEMENT]

    elif isinstance(value, list):
        return list(unwrap(item) for item in value)
    else:
        return value


def ensure_task(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        return await asyncio.get_event_loop().create_task(func(*args, **kwargs))

    return wrapper


def strip_auth(url: str) -> str:
    pr = urlparse(url)
    safe_netloc = pr.hostname
    if pr.port:
        safe_netloc = f"{safe_netloc}:{pr.port}"
    return urlunparse(
        (pr.scheme, safe_netloc, pr.path, pr.params, pr.query, pr.fragment)
    )


def check_response_error(*, status: int, data: Any) -> None:
    if status >= 400:
        errors.raise_exception(data, status)
    if not isinstance(data, dict):
        return
    data_status = data.get("status", None)
    if data_status is None:
        return
    if data_status == constants.STATUS_SUCCESS:
        return
    errors.raise_exception(data, status)


class Connection:
    def __init__(self, session: httpx.AsyncClient, prefix: str):
        self.session = session
        self.prefix = prefix

    @ensure_task
    async def request(
        self, *, url: str, method: str, data=None, timeout=None
    ) -> Tuple[int, Any]:
        header = {"Content-Type": "application/json"}
        if data is None:
            data = {}
        if method not in {"POST", "PUT"}:
            data = None
            header = None

        body = json.dumps(data) if data is not None else None
        full_url = self.prefix + url

        log.info("→ Request %s %s", method, strip_auth(full_url))
        if header or body:
            log.debug("Headers: %s | Body: %s", header, body)

        response = await self.session.request(
            url=full_url, method=method, headers=header, content=body, timeout=timeout
        )
        response_body = await response.aread()

        try:
            data = json.loads(response_body)
        except JSONDecodeError as exc:
            log.error("Failed to decode JSON from response: %s", exc)
            log.debug("Response body: %s", response_body)
            data = {"error": "!internal", "message": str(exc), "stacktrace": ""}

        wrap_screen(data)

        log.info(
            "← Response %s %s (%d)", method, strip_auth(full_url), response.status_code
        )
        log.debug("Response data: %s", data)

        check_response_error(data=data, status=response.status_code)
        return response.status_code, data

    async def upload_file(self, path: Path) -> Path:
        log.info("Uploading local file: %s", path)
        return path

    def prefixed(self, prefix: str) -> "Connection":
        return self.__class__(self.session, self.prefix + prefix)


class RemoteConnection(Connection):
    async def upload_file(self, path: Path) -> Path:
        fobj = BytesIO()
        with ZipFile(fobj, "w", ZIP_DEFLATED) as zf:
            zf.write(path, path.name)

        content = base64.b64encode(fobj.getvalue()).decode("utf-8")
        status, data = await self.request(
            url="/file", method="POST", data={"file": content}
        )

        value = unwrap(data.get("value", None))
        log.info("File uploaded successfully: %s → %s", path, value)
        return Path(value)



import abc
import asyncio
import re
import sys
from dataclasses import dataclass
from functools import partial
from typing import List, TextIO, Optional

import httpx
from packaging import version

from wapiti_arsenic.connection import Connection, RemoteConnection
from wapiti_arsenic.errors import ArsenicError
from wapiti_arsenic.http import Auth, BasicAuth
from wapiti_arsenic.subprocess import get_subprocess_impl
from wapiti_arsenic.utils import free_port
from wapiti_arsenic.webdriver import WebDriver


async def tasked(coro):
    return await asyncio.get_event_loop().create_task(coro)


async def check_service_status(client: httpx.AsyncClient, url: str) -> bool:
    response = await client.get(url + "/status")
    return 200 <= response.status_code < 300


async def subprocess_based_service(
    cmd: List[str],
    service_url: str,
    log_file: TextIO,
    start_timeout: float = 15,
) -> WebDriver:
    closers = []
    try:
        impl = get_subprocess_impl()
        process = await impl.start_process(cmd, log_file)
        closers.append(partial(impl.stop_process, process))
        client = httpx.AsyncClient()
        closers.append(client.aclose)

        async def wait_service():
            # Wait for service with exponential back-off
            for i in range(-10, 9999):
                try:
                    ok = await tasked(check_service_status(client, service_url))
                except httpx.ConnectError:
                    # We possibly checked too quickly
                    ok = False
                if ok:
                    return
                await asyncio.sleep(start_timeout * 2**i)

        try:
            await asyncio.wait_for(wait_service(), timeout=start_timeout)
        except asyncio.TimeoutError:
            raise ArsenicError("not starting?")
        return WebDriver(Connection(client, service_url), closers)
    except:
        for closer in reversed(closers):
            await closer()
        raise


class Service(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def start(self) -> WebDriver:
        raise NotImplementedError()


@dataclass
class Geckodriver(Service):
    log_file: TextIO = sys.stdout
    binary: str = "geckodriver"
    version_check: bool = True
    start_timeout: float = 15

    _version_re = re.compile(r"geckodriver (\d+\.\d+)")

    async def _check_version(self):
        if self.version_check:
            impl = get_subprocess_impl()
            output = await impl.run_process([self.binary, "--version"])
            match = self._version_re.search(output)
            if not match:
                raise ValueError(
                    "Could not determine version of geckodriver. To "
                    "disable version checking, set `version_check` to "
                    "`False`."
                )
            binary_version_str = match.group(1)
            binary_version = version.parse(binary_version_str)
            if binary_version < version.parse("0.16.1"):
                raise ValueError(
                    f"Geckodriver version {binary_version_str} is too old. 0.16.1 or "
                    f"higher is required. To disable version checking, set "
                    f"`version_check` to `False`."
                )

    async def start(self):
        port = free_port()
        await self._check_version()
        return await subprocess_based_service(
            [self.binary, "--port", str(port)],
            f"http://localhost:{port}",
            self.log_file,
            start_timeout=self.start_timeout,
        )


@dataclass
class Chromedriver(Service):
    log_file: TextIO = sys.stdout
    binary: str = "chromedriver"
    start_timeout: float = 15

    async def start(self):
        port = free_port()
        return await subprocess_based_service(
            [self.binary, f"--port={port}"],
            f"http://localhost:{port}",
            self.log_file,
            start_timeout=self.start_timeout,
        )


@dataclass
class MSEdgeDriver(Service):
    log_file: TextIO = sys.stdout
    binary: str = "msedgedriver"
    start_timeout: float = 15

    async def start(self):
        port = free_port()
        return await subprocess_based_service(
            [self.binary, f"--port={port}"],
            f"http://localhost:{port}",
            self.log_file,
            start_timeout=self.start_timeout,
        )


def auth_or_string(value):
    if value is None:
        return value

    if isinstance(value, Auth):
        return value

    if isinstance(value, str) and value.count(":") == 1:
        username, password = value.split(":")
        return BasicAuth(username, password)

    raise TypeError()


@dataclass
class Remote(Service):
    url: str
    auth: Optional[Auth] = None

    def __post_init__(self):
        self.auth = auth_or_string(self.auth)

    async def start(self):
        closers = []
        headers = {}
        if self.auth:
            headers.update(self.auth.get_headers())
        try:
            client = httpx.AsyncClient(headers=headers)
            closers.append(client.aclose)
            return WebDriver(RemoteConnection(client, self.url), closers)
        except:
            for closer in reversed(closers):
                await closer()
            raise


@dataclass
class IEDriverServer(Service):
    log_file: TextIO = sys.stdout
    binary: str = "IEDriverServer.exe"
    log_level: str = "FATAL"
    start_timeout: float = 15

    async def start(self):
        port = free_port()
        return await subprocess_based_service(
            [self.binary, f"/port={port}", f"/log-level={self.log_level}"],
            f"http://localhost:{port}",
            self.log_file,
            start_timeout=self.start_timeout,
        )



import base64
from dataclasses import dataclass
from functools import partial
from io import BytesIO
from pathlib import Path
from typing import Awaitable, Callable, Any, List, Dict, Tuple, Iterator

from wapiti_arsenic import constants
from wapiti_arsenic.connection import Connection, unwrap
from wapiti_arsenic.constants import SelectorType, WindowType
from wapiti_arsenic.errors import NoSuchElement, OperationNotSupported
from wapiti_arsenic.utils import Rect

UNSET = object()


def escape_value(value: str) -> str:
    if '"' in value and "'" in value:
        parts = value.split('"')
        result = ["concat("]
        for part in parts:
            result.append(f'"{part}"')
            result.append(", '\"', ")

        result = result[0:-1]
        if value.endswith('"'):
            return "".join(result) + ")"

        return "".join(result[:-1]) + ")"

    if '"' in value:
        return f"'{value}'"

    return f'"{value}"'


class RequestHelpers:
    async def _request(
        self, *, url: str, method: str, data=None, raw=False, timeout=None
    ):
        status, data = await self.connection.request(
            url=url, method=method, data=data, timeout=timeout
        )
        if raw:
            return data
        if data:
            return self._unwrap(data.get("value", None))

    def _unwrap(self, value):
        """
        Unwrap a value returned from a webdriver. Specifically this means trying to
        extract the element ID of a web element or a list of web elements.
        """
        return unwrap(value)


class Element(RequestHelpers):
    def __init__(self, id: str, connection: Connection, session: "Session"):
        self.id = id
        self.connection = connection
        self.session = session

    async def get_text(self) -> str:
        return await self._request(url="/text", method="GET")

    async def send_keys(self, keys: str):
        await self._request(
            url="/value", method="POST", data={"value": list(keys), "text": keys}
        )

    async def send_file(self, path: Path):
        path = await self.session.connection.upload_file(path)
        await self.send_keys(str(path))

    async def clear(self):
        await self._request(url="/clear", method="POST")

    async def click(self):
        await self._request(url="/click", method="POST")

    async def is_displayed(self) -> bool:
        return await self._request(url="/displayed", method="GET")

    async def is_enabled(self) -> bool:
        return await self._request(url="/enabled", method="GET")

    async def get_attribute(self, name: str) -> str:
        return await self._request(url=f"/attribute/{name}", method="GET")

    async def get_property(self, name: str) -> str:
        return await self._request(url=f"/property/{name}", method="GET")

    async def get_css_value(self, name: str) -> str:
        return await self._request(url=f"/css/{name}", method="GET")

    async def select_by_value(self, value: str):
        value = escape_value(value)
        option = await self.get_element(f"option[value={value}]")
        await option.click()

    async def get_element(
        self, selector: str, selector_type: SelectorType = SelectorType.css_selector
    ) -> "Element":
        element_id = await self._request(
            url="/element",
            method="POST",
            data={"using": selector_type.value, "value": selector},
        )
        return self.session.create_element(element_id)

    async def get_elements(
        self, selector: str, selector_type: SelectorType = SelectorType.css_selector
    ) -> List["Element"]:
        element_ids = await self._request(
            url="/elements",
            method="POST",
            data={"using": selector_type.value, "value": selector},
        )
        return [self.session.create_element(element_id) for element_id in element_ids]

    async def get_rect(self):
        data = await self._request(url="/rect", method="GET")
        return Rect(data["x"], data["y"], data["width"], data["height"])

    async def get_screenshot(self):
        return BytesIO(
            base64.b64decode(await self._request(url="/screenshot", method="GET"))
        )


TCallback = Callable[..., Awaitable[Any]]
TWaiter = Callable[[int, TCallback], Awaitable[Any]]


class Session(RequestHelpers):
    element_class = Element

    def __init__(
        self, connection: Connection, wait: TWaiter, driver, browser, bind: str = ""
    ):
        self.connection = connection
        self.bind = bind
        self.wait = wait
        self.driver = driver
        self.browser = browser

    async def request(
        self, url: str, method: str = "GET", data: Dict[str, Any] = UNSET
    ):
        return await self._request(url=url, method=method, data=data)

    async def get(self, url: str, timeout=None):
        await self._request(
            url="/url", method="POST", data={"url": self.bind + url}, timeout=timeout
        )

    async def get_url(self):
        return await self._request(url="/url", method="GET")

    async def get_page_source(self) -> str:
        return await self._request(url="/source", method="GET")

    async def get_element(
        self, selector: str, selector_type: SelectorType = SelectorType.css_selector
    ) -> Element:
        element_id = await self._request(
            url="/element",
            method="POST",
            data={"using": selector_type.value, "value": selector},
        )
        return self.create_element(element_id)

    async def get_elements(
        self, selector: str, selector_type: SelectorType = SelectorType.css_selector
    ) -> List[Element]:
        result = await self._request(
            url="/elements",
            method="POST",
            data={"using": selector_type.value, "value": selector},
        )
        return [self.create_element(element_id) for element_id in result]

    async def wait_for_element(
        self,
        timeout: int,
        selector: str,
        selector_type: SelectorType = SelectorType.css_selector,
    ) -> Element:
        return await self.wait(
            timeout, partial(self.get_element, selector, selector_type), NoSuchElement
        )

    async def wait_for_element_gone(
        self,
        timeout: int,
        selector: str,
        selector_type: SelectorType = SelectorType.css_selector,
    ):
        async def callback():
            try:
                await self.get_element(selector, selector_type)
            except NoSuchElement:
                return True

            return False

        return await self.wait(timeout, callback)

    async def add_cookie(
        self,
        name: str,
        value: str,
        *,
        path: str = UNSET,
        domain: str = UNSET,
        secure: bool = UNSET,
        expiry: int = UNSET,
        httponly: bool = UNSET,
    ):
        cookie = {"name": name, "value": value}
        if path is not UNSET:
            cookie["path"] = path
        if domain is not UNSET:
            cookie["domain"] = domain
        if secure is not UNSET:
            cookie["secure"] = secure
        if expiry is not UNSET:
            cookie["expiry"] = expiry
        if httponly is not UNSET:
            cookie["HttpOnly"] = httponly
        await self._request(url="/cookie", method="POST", data={"cookie": cookie})

    async def get_cookie(self, name: str):
        return await self._request(url=f"/cookie/{name}", method="GET")

    async def get_all_cookies(self):
        return await self._request(url="/cookie", method="GET")

    async def delete_cookie(self, name: str):
        await self._request(url=f"/cookie/{name}", method="DELETE")

    async def delete_all_cookies(self):
        await self._request(url="/cookie", method="DELETE")

    async def execute_script(self, script: str, *args: Any):
        return await self._request(
            url="/execute/sync",
            method="POST",
            data={"script": script, "args": list(args)},
        )

    async def execute_async_script(self, script: str, *args: Any):
        return await self._request(
            url="/execute/async",
            method="POST",
            data={"script": script, "args": list(args)},
        )

    async def set_window_size(self, width: int, height: int, handle: str = "current"):
        return await self._request(
            url="/window/rect",
            method="POST",
            data={"width": width, "height": height, "windowHandle": handle},
        )

    async def get_window_size(self, handle: str = "current") -> Tuple[int, int]:
        return await self._request(
            url="/window/rect", method="GET", data={"windowHandle": handle}
        )

    async def set_window_fullscreen(self, handle: str = "current"):
        return await self._request(
            url="/window/fullscreen", method="POST", data={"windowHandle": handle}
        )

    async def set_window_maximize(self, handle: str = "current"):
        return await self._request(
            url="/window/maximize", method="POST", data={"windowHandle": handle}
        )

    async def get_alert_text(self) -> str:
        return await self._request(url="/alert/text", method="GET")

    async def send_alert_text(self, value: str):
        return await self._request(
            url="/alert/text", method="POST", data={"text": value}
        )

    async def dismiss_alert(self):
        return await self._request(url="/alert/dismiss", method="POST")

    async def accept_alert(self):
        return await self._request(url="/alert/accept", method="POST")

    async def perform_actions(self, actions: Dict[str, Any]):
        return await self._request(url="/actions", method="POST", data=actions)

    async def get_screenshot(self) -> BytesIO:
        return BytesIO(
            base64.b64decode(await self._request(url="/screenshot", method="GET"))
        )

    async def close(self):
        await self._request(url="", method="DELETE")

    def create_element(self, element_id):
        return self.element_class(
            element_id, self.connection.prefixed(f"/element/{element_id}"), self
        )

    async def get_window_handles(self):
        return await self._request(url="/window/handles", method="GET")

    async def get_window_handle(self):
        return await self._request(url="/window", method="GET")

    async def switch_to_window(self, handle):
        return await self._request(
            url="/window", method="POST", data={"handle": handle, "name": handle}
        )

    async def new_window(self, window_type: WindowType = WindowType.tab):
        return await self._request(
            url="/window/new", method="POST", data={"type": window_type.value}
        )


def _pointer_down(device, action):
    del action["duration"]
    url = (
        "/buttondown"
        if device["parameters"]["pointerType"] == "mouse"
        else "/touch/down"
    )
    return url, "POST", action


def _pointer_up(device, action):
    del action["duration"]
    url = "/buttonup" if device["parameters"]["pointerType"] == "mouse" else "/touch/up"
    return url, "POST", action


def _pointer_move(device, action):
    del action["duration"]
    url = "/moveto" if device["parameters"]["pointerType"] == "mouse" else "/touch/move"
    origin = action["origin"]
    if origin == "pointer":
        data = {"xoffset": action["x"], "yoffset": action["y"]}
    elif constants.WEB_ELEMENT in origin:
        data = {"element": origin[constants.WEB_ELEMENT]}
    else:
        raise OperationNotSupported(f"Cannot move using origin {origin}")
    return url, "POST", data


def _pause(device, action):
    return None


def key_down(device, action):
    return "/keydown", "POST", {""}


legacy_actions = {
    ("pointer", "pointerDown"): _pointer_down,
    ("pointer", "pointerUp"): _pointer_up,
    ("pointer", "pointerMove"): _pointer_move,
    ("pointer", "pause"): _pause,
    ("key", "pause"): _pause,
}


@dataclass
class LegacyAction:
    device: Dict[str, Any]
    action: Dict[str, Any]


def get_legacy_actions(devices: List[Dict[str, Any]]) -> Iterator[LegacyAction]:
    i = 0
    while devices:
        for device in devices:
            action = device["actions"].pop(0)
            i += 1
            yield LegacyAction(device, action)
        devices = [device for device in devices if device["actions"]]


def transform_legacy_actions(
    devices: List[Dict[str, Any]],
) -> Iterator[Tuple[str, str, Dict[str, Any]]]:
    for legacy_action in get_legacy_actions(devices):
        device_type = legacy_action.device["type"]
        action_type = legacy_action.action["type"]
        device = {
            key: value for key, value in legacy_action.device.items() if key != "type"
        }
        action = {
            key: value for key, value in legacy_action.action.items() if key != "type"
        }
        try:
            handler = legacy_actions[(device_type, action_type)]
        except KeyError:
            raise OperationNotSupported(
                f"Unsupported action {action_type} for device_type {device_type}"
            )
        action = handler(device, action)
        if action is not None:
            yield action



import asyncio
from typing import Awaitable, Callable, List, Any, Union, Type

import time

from wapiti_arsenic.browsers import Browser
from wapiti_arsenic.connection import Connection
from wapiti_arsenic.errors import ArsenicError, ArsenicTimeout, SessionStartError
from wapiti_arsenic.session import Session


class SessionContext:
    def __init__(self, driver: "WebDriver", browser: Browser, bind: str):
        self.driver = driver
        self.browser = browser
        self.bind = bind
        self.session: Session = None

    async def __aenter__(self) -> Session:
        self.session = await self.driver.new_session(self.browser, self.bind)
        return self.session

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.session.close()
        self.session = None


TClosers = List[Callable[..., Awaitable[None]]]


class WebDriver:
    def __init__(self, connection: Connection, closers: TClosers):
        self.connection = connection
        self.closers = closers

    def session(self, browser: Browser, bind="") -> SessionContext:
        return SessionContext(self, browser, bind)

    async def new_session(self, browser: Browser, bind="") -> Session:
        status, response = await self.connection.request(
            url="/session",
            method="POST",
            data={"capabilities": {"alwaysMatch": browser.capabilities}},
        )
        original_response = response
        if "sessionId" not in response:
            response = response["value"]
        if "sessionId" not in response:
            if "error" in original_response:
                err_resp = original_response
            elif "error" in response:
                err_resp = response
            else:
                raise SessionStartError("Unknown", "Unknown", original_response)
            raise SessionStartError(
                err_resp["error"], err_resp.get("message", ""), original_response
            )
        session_id = response["sessionId"]
        return browser.session_class(
            connection=self.connection.prefixed(f"/session/{session_id}"),
            bind=bind,
            wait=self.wait,
            driver=self,
            browser=browser,
        )

    async def close(self):
        for closer in reversed(self.closers):
            await closer()

    async def wait(
        self,
        timeout: Union[float, int],
        func: Callable[[], Awaitable[Any]],
        *exceptions: Exception,
    ) -> Any:
        deadline = time.time() + timeout
        err = None
        while deadline > time.time():
            try:
                result = await func()
                if result:
                    return result
                else:
                    await asyncio.sleep(0.2)
            except exceptions as exc:
                err = exc
                await asyncio.sleep(0.2)
        raise ArsenicTimeout() from err



"""
This module contains special keys as defined in the web driver specification.

They can be used in the send_keys API using string formatting.
"""

NULL = "\ue000"
CANCEL = "\ue001"
HELP = "\ue002"
BACKSPACE = "\ue003"
TAB = "\ue004"
CLEAR = "\ue005"
RETURN = "\ue006"
ENTER = "\ue007"
SHIFT = "\ue008"
CONTROL = "\ue009"
ALT = "\ue00a"
PAUSE = "\ue00b"
ESCAPE = "\ue00c"
SPACE = "\ue00d"
PAGE_UP = "\ue00e"
PAGE_DOWN = "\ue00f"
END = "\ue010"
HOME = "\ue011"
LEFT = "\ue012"
UP = "\ue013"
RIGHT = "\ue014"
DOWN = "\ue015"
INSERT = "\ue016"
DELETE = "\ue017"
SEMICOLON = "\ue018"
EQUALS = "\ue019"

NUMPAD0 = "\ue01a"
NUMPAD1 = "\ue01b"
NUMPAD2 = "\ue01c"
NUMPAD3 = "\ue01d"
NUMPAD4 = "\ue01e"
NUMPAD5 = "\ue01f"
NUMPAD6 = "\ue020"
NUMPAD7 = "\ue021"
NUMPAD8 = "\ue022"
NUMPAD9 = "\ue023"
MULTIPLY = "\ue024"
ADD = "\ue025"
SEPARATOR = "\ue026"
SUBTRACT = "\ue027"
DECIMAL = "\ue028"
DIVIDE = "\ue029"

F1 = "\ue031"
F2 = "\ue032"
F3 = "\ue033"
F4 = "\ue034"
F5 = "\ue035"
F6 = "\ue036"
F7 = "\ue037"
F8 = "\ue038"
F9 = "\ue039"
F10 = "\ue03a"
F11 = "\ue03b"
F12 = "\ue03c"

META = "\ue03d"
COMMAND = "\ue03d"
